//
// GameX - Input-related WindowsDX Class Code
// 
// Copyright (C) 2002 Rama C. Hoetzlein
// 
// This software is released under the GameX GNU GPL 
// Open Source License. See the GameX documentation included
// with this source code for terms of modification, 
// distribution and re-release. 
// 
// this is a version of GameX that has been modified (by Justin Pease)


#include "gamex-win-dx.hpp"




bool WindowsDX::UpdateInput (void)
{
	bool skip_keyboard = (di_keyboard == NULL);
	bool skip_mouse = (di_mouse == NULL) || win_request_no_mouse_input;

	if(!skip_keyboard) {
		if(FAILED(di_keyboard->GetDeviceState(256, (LPVOID)di_keystate))) {
			if(SUCCEEDED(di_keyboard->Acquire())) {
				if(FAILED(di_keyboard->GetDeviceState(256, (LPVOID)di_keystate))) {
					skip_keyboard = true;
				}
			}
		}

		if(!skip_keyboard) {

			{
				for(KeyID k = 0 ; k < 255 ; k++) {
					if(di_keypressed[k] & INPUT_PRESSED) // if just pressed last time,
						di_keypressed[k] ^= (INPUT_HELD|INPUT_PRESSED); // switch pressed off, held on
				}
			}

			// fix windows oddity that makes the right shift button register as unpressed
			// if an arrow key is pressed while it's down
			if(!IsKeyDown(KEY_RSHIFT) && (GetAsyncKeyState(VK_RSHIFT) & 0x8000))
				di_keystate[KEY_RSHIFT] |= INPUT_PRESSED;

			// transfer/record unique (non-held) keypresses
			for(KeyID k = 0 ; k < 255 ; k++) {
				if(di_keypressed[k] & INPUT_HELD) {
					if(!di_keystate[k])
						di_keypressed[k] = di_keystate[k];
				} else {
					di_keypressed[k] = di_keystate[k];
				}
			}

			if(IsEscapeReserved() && GetKeyPress(KEY_ESCAPE) // ESC is reserved to quit unless RUN_USEESCAPEKEY is on
			|| IsCtrlDown() && GetKeyPress(KEY_Q) // whenever CTRL is down, Q is reserved to quit (nonstandard but intuitive)
			|| IsAltDown() && GetKeyPress(KEY_F4)) // whenever ALT is down, F4 is reserved to quit (faster than in WindowFunc)
				Quit();


			if(IsAltEnterReserved() && GetKeyPress(KEY_ENTER) && IsAltDown()) // ALT+ENTER is reserved to toggle fullscreen mode unless RUN_NOALTENTERTOGGLE is on
				GameX.ToggleFullscreen();
		}
	}

	if(!skip_mouse) {

		if(FAILED(di_mouse->GetDeviceState(sizeof(DIMOUSESTATE), (LPVOID)&di_mousestate))) {
			if(SUCCEEDED(di_mouse->Acquire())) {
				if(FAILED(di_mouse->GetDeviceState(sizeof(DIMOUSESTATE), (LPVOID)&di_mousestate))) {
					skip_mouse = true;
				}
			}
		}

		if(!skip_mouse) {

			{
				for(MouseButtonID b = 0 ; b < 4 ; b++) {
					if(di_mousepressed[b] & INPUT_PRESSED) // if just pressed last time,
						di_mousepressed[b] ^= (INPUT_HELD|INPUT_PRESSED); // switch pressed off, held on
				}
			}
			
			// if the mouse isn't outside of the game,
			if(win_fullscreen || (mouse_x>0 && mouse_y>0 && GetMouseX()<win_width && GetMouseY()<win_height)) {
				// transfer/record unique (non-held) mousepresses
				for(MouseButtonID b = 0 ; b < 4 ; b++) {
					if(!di_mousestate.rgbButtons[b])
						mouse_down[b] = 0;
					else if(!mouse_down[b])
		/// 			di_mousestate.rgbButtons[b] = 0; // seems to cause mouse click detection to lag after typing
						;
					else if(mouse_down[b] == 2 && di_mousestate.rgbButtons[b])
						di_mousestate.rgbButtons[b] |= 0x01;
					if(di_mousepressed[b] & INPUT_HELD) {
						if(!di_mousestate.rgbButtons[b])
							di_mousepressed[b] = di_keystate[b];
					} else {
						di_mousepressed[b] = di_mousestate.rgbButtons[b];
					}
				}
			}
			if(di_mousestop) {
				di_mousestate.lX = di_mousestate.lY = di_mousestate.lZ = 0;
				if(win_active)
					di_mousestop--;

				// unregister all clicks at this time (so the game doesn't respond to window-activating clicks)
				for(MouseButtonID b = 0 ; b < 4 ; b++) {
					mouse_down[b] = 0;
					di_mousepressed[b] = 0;
				}
			}

			POINT mousePoint;
			GetCursorPos(&mousePoint);
			mousePoint.x -= win_xpos+win_xoffset;
			mousePoint.y -= win_ypos+win_yoffset;
			SetMouseMove(mousePoint.x,mousePoint.y);
		}
	}

	return true;
}

bool WindowsDX::AnyKeyPress (void)	
{
	for(int n=0; n<256; n++)
		if(GetKeyPress(n))
			// some keys don't count as "any key", such as volume control keys
			if(n != DIK_CAPITAL && n != DIK_MUTE && n != DIK_VOLUMEDOWN && n != DIK_VOLUMEUP && n != DIK_SYSRQ
			&& n != DIK_POWER && n != DIK_SLEEP && n != DIK_WAKE && n != DIK_NUMLOCK && n != DIK_SCROLL
			&& n != DIK_NEXTTRACK && n != DIK_WEBSEARCH && n != DIK_WEBFAVORITES && n != DIK_WEBREFRESH && n != DIK_WEBSTOP
			&& n != DIK_WEBHOME && n != DIK_WEBFORWARD && n != DIK_WEBBACK && n != DIK_MYCOMPUTER && n != DIK_MAIL
			&& n != DIK_MEDIASELECT && n != DIK_PLAYPAUSE && n != DIK_MEDIASTOP && n != DIK_CALCULATOR)
				return true;
	return false;
}

bool WindowsDX::PeekAnyKeyPress (void)	
{
	for(int n=0; n<256; n++)
		if(PeekKeyPress(n))
			// some keys don't count as "any key", such as volume control keys
			if(n != DIK_CAPITAL && n != DIK_MUTE && n != DIK_VOLUMEDOWN && n != DIK_VOLUMEUP && n != DIK_SYSRQ
			&& n != DIK_POWER && n != DIK_SLEEP && n != DIK_WAKE && n != DIK_NUMLOCK && n != DIK_SCROLL
			&& n != DIK_NEXTTRACK && n != DIK_WEBSEARCH && n != DIK_WEBFAVORITES && n != DIK_WEBREFRESH && n != DIK_WEBSTOP
			&& n != DIK_WEBHOME && n != DIK_WEBFORWARD && n != DIK_WEBBACK && n != DIK_MYCOMPUTER && n != DIK_MAIL
			&& n != DIK_MEDIASELECT && n != DIK_PLAYPAUSE && n != DIK_MEDIASTOP && n != DIK_CALCULATOR)
				return true;
	return false;
}