//
// GameX - Main-Loop-related WindowsDX Class Code (including WinMain and WinProc (WindowFunc()))
// 
// Copyright (C) 2002 Rama C. Hoetzlein
// 
// This software is released under the GameX GNU GPL 
// Open Source License. See the GameX documentation included
// with this source code for terms of modification, 
// distribution and re-release. 
// 
// this is a version of GameX that has been modified (by Justin Pease)


// KNOWN ISSUES:

// In windowed mode, if you click out of a game, then click once on the game's title bar to get back into the game,
// Windows forgets to visibly activate the game window even though GameX still gets the activation message,
// resulting in the game playing sort-of in the background until it's clicked on again.
// May not be universal across computers or specific to GameX, however.

#include "gamex-win-dx.hpp"

#include "gamex-win-dx-timer.inl"

// WinMain and WindowFunc functions supplied by WindowsDX
LRESULT CALLBACK WindowFunc (HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch (message) {
	case WM_MOVE: // window movement detection (non-windowed mode)
		GameX.SetPosition (LOWORD(lParam), HIWORD(lParam)) ;
		GameX.SetMouseStop(1); // so window movement doesn't register as mouse movement to the game
		break;
	case WM_CHAR: // buffered keyboard input detection, for GetKey()
		GameX.AddBufferedKeyPress ((TCHAR) wParam) ; // add key char to buffer
		return 0;
	case WM_SYSKEYDOWN: // detects pressing of system keys such as ALT
		if(wParam == VK_MENU)
			return 0; // prevent pressing the alt key on its own from pausing/freezing the game
		break;		  // (returning 0 stops Windows from taking action, breaking lets it do what it wants)
	case WM_SYSCOMMAND: // detects system commands such as closing the window, etc.
		switch(wParam & 0xFFF0) {
			case SC_CLOSE:
				if(!GameX.IsQuitDialogOpen())
					GameX.Quit();
				return 0;
			case SC_MONITORPOWER:
				if(GameX.IsFullScreen())
					return 0;
			case SC_KEYMENU:
			case SC_PREVWINDOW:
				return 0; // prevents ALT-anykey from blinking the window and beeping
						  // and prevents ALT-F6 from pausing the game
						  // but allows ALT-F4 to continue functioning as quit button
		}
		// no break
	case WM_ACTIVATEAPP: // game switched to
		// no break 	  // or
	case WM_ENTERMENULOOP: // menu opened
		if(GameX.CanDraw() && GameX.CanAutoPause() && !GameX.IsQuitDialogOpen()) {
#ifndef GAMEX_RUN_UNMANAGED
			GameX.SetWarningOnDraw(false);
			GameDraw();
#endif											// to indicate game being paused because it is in the background 
			GameX.win_alternator = 0;						// or has a menu or system command being accessed,
			GameX.SpecialPauseFade(0.125f,0.25f,0,0,32,64,64,128,160); // partially fade the screen (blue-ish)
		}
		GameX.UpdateView() ;
		break;
	case WM_KILLFOCUS: // window deactivated
		if(GameX.win_active) {
			GameX.SetActivation(false) ;
			if(GameX.CanAutoPause()) {
				GameX.PauseMusic();
				char title_str [256];
				sprintf(title_str, "Paused: %s - GameX 5.0", GameX.GetName());
				SetWindowText(GameX.GetWindow(),title_str);
			}
			if(GameX.CanHideCursor()) GameX.UnhideCursor();
			if(GameX.CanAutoPause()) 
			GameX.SetMouseStop(5); // skip 5 mouse input loops, the mouse should be ready for input by then
		}
		break;
	case WM_SETFOCUS: // window activated
		if(!GameX.win_active && GameX.win_realy>1) {
			GameX.SetActivation (true) ;
			if(GameX.CanAutoPause()) {
				GameX.ResumeMusic();
				char title_str [256];
				sprintf(title_str, "%s - GameX 5.0", GameX.GetName());
				SetWindowText(GameX.GetWindow(),title_str);
			}
			if(GameX.CanHideCursor())
				GameX.HideCursor();
		}
		break;
	case WM_PAINT: // redraw of game screen requested
		GameX.UpdateView(); // (in addition to the frequent timed updates, in case game is suspended)
		break;
	case WM_CLOSE: // close box or command chosen
		if(GameX.CheckQuit()) {
			GameX.SetStatus(GAMEX_SHUTTINGDOWN);
			DestroyWindow (hWnd) ; // quit the game by activating WM_DESTROY case next
		}
		else
			return 0; // user decided against quitting, continue game
		break;
	case WM_DESTROY:
		PostQuitMessage(0) ; // post the message that actually quits the application/game
		break;

	// these are used to check to make sure DirectInput's mouse presses are within the window
	case WM_LBUTTONDOWN:	GameX.SetMousePress(MOUSE_LEFT);			return 0;
	case WM_MBUTTONDOWN:	GameX.SetMousePress(MOUSE_MIDDLE);			return 0;
	case WM_RBUTTONDOWN:	GameX.SetMousePress(MOUSE_RIGHT);			return 0;

	// these are used to allow double-click detection (only works if CS_DBLCLKS is set on our WNDCLASS.style)
	case WM_LBUTTONDBLCLK:	GameX.SetMouseDoubleClick(MOUSE_LEFT);		return 0;
	case WM_MBUTTONDBLCLK:	GameX.SetMouseDoubleClick(MOUSE_MIDDLE);	return 0;
	case WM_RBUTTONDBLCLK:	GameX.SetMouseDoubleClick(MOUSE_RIGHT); 	return 0;

	case WM_SIZE: // user resizes the game window in windowed mode
		if(GameX.GetStatus() == GAMEX_READY) {
			if(HIWORD(lParam) < 1) {
				PostMessage(hWnd,WM_KILLFOCUS,0,0);
			} else {
				PostMessage(hWnd,WM_SETFOCUS,0,0);
			}

			if(!GameX.IsResizeWorkaroundEnabled()) {
				GameX.win_realx = LOWORD(lParam);
				GameX.win_realy = HIWORD(lParam);
				GameX.SetProportional();
			} else {
				GameX.win_xoffset = (LOWORD(lParam) - GameX.GetWidth()) / 2;
				GameX.win_yoffset = (HIWORD(lParam) - GameX.GetHeight()) / 2;				
				GameX.win_realx = GameX.GetWidth();
				GameX.win_realy = GameX.GetHeight();
			}

			if(GameX.win_realy < 1) GameX.win_realy = 1; // avoids crash
			if(GameX.win_realx < 1) GameX.win_realx = 1;
			GameX.SetMouseStop(1); // so window resize doesn't register as mouse movement to the game
		}
		break;
	case WM_DIRECTSHOW_EVENT:
		GameX.HandleDirectShowEvent();
		break;
	case WM_KEYUP:
		if(wParam == VK_SNAPSHOT) { // ("Prnt Scrn" button released)
		#ifndef GAMEX_RUN_UNMANAGED
			if(GameX.CanShowRunInfo()) {GameX.SetWarningOnDraw(false); GameDraw();} // clear debug info if it's on
		#endif
			ImageExt ie; ie.Save("screenshot.bmp",VIEWPORT); // make taking screenshots easier
			GameX.CopyToClipboard(VIEWPORT); // copy back buffer directly to clipboard as well as to file
			return 0;
		}
		break;
	}
	
	return DefWindowProc (hWnd, message, wParam, lParam) ;		
}




int WINAPI WinMain (HINSTANCE hThisInst, HINSTANCE hPrevInst, LPSTR lpszArgs, int nWinMode)
{
	MSG msg;
	
	GameX.InitWindowsInstance (hThisInst, nWinMode);

	GameInit() ; // defined by game, hopefully calls GameX.Initialize()

	if (GameX.GetStatus() != GAMEX_READY) {
		GameX.ReportProblem() ;
		GameX.SetStatus(GAMEX_READY); // after temp auto-init
	} else {
		GameX.DisplayLoadingWithValue (90);
//		Sleep(750); // wait until async parts of DX init should be complete, to avoid pause on 1st frame
		GameX.DisplayLoadingWithValue (100);
//		Sleep(750); // (total wait of 1500 milliseconds is sometimes necessary)
	}
if(GameX.win_log_debug) {
	debug.Output("-- Graphic/Sound memory in use after GameInit(): %d KB\n", GameX.GetGraphicMemUsed());
	debug.Output("WinDX: Running Game...\n");
}

	int frame=0, cycle=0; // so we can keep track of the frame rate
	GameX.sleep_record = 0; // to keep track of how much time we spend sleeping

	GameX.win_global_counter = 0;
	GameX.win_global_frame_total = 0;
	hyper cps = GameX.win_cps; // cycles per second (a cycle meaning a call to GameRun(), not a CPU cycle)
	hyper fps = GameX.win_fps; // frames per second (a frame meaning a call to GameDraw() and UpdateView())
	hyper tpc = GetClocksPerSec() / (cps?cps:1); // time per cycle (number of clock ticks between each game update)
	hyper tpf = GetClocksPerSec() / (fps?fps:1); // time per frame (number of clock ticks between each view update)
	bool needUpdate = false; // (start as false because GameRun() should be called first before GameDraw())
	bool update_auto_status = true;
	hyper cycleTime = GetClock(); // timer for game update
	hyper frameTime = GetClock(); // timer for view update
	hyper frameClock = GetClock(); // timer for framerate calculation update
	int leftover = 0;
	for(;;) {
		if(PeekMessage (&msg, NULL, 0, 0, PM_REMOVE)) {
			if (msg.message == WM_QUIT) break;
			TranslateMessage(&msg) ;
			DispatchMessage(&msg) ;
		}
		if(!GameX.win_active && GameX.CanAutoPause()) { // if the window isn't active, and we can pause,
			Sleep(10) ; // sleep so we don't max out the cpu in the background
			cycleTime = GetClock(); // set the timers to the current time
			frameTime = GetClock(); // so we don't skip a ton of frames when resuming
		}
		else {

		#ifndef GAMEX_RUN_UNMANAGED

			// AUTOMATIC GAME SPEED AND FRAME RATE MANAGEMENT:
			// This ensures that the game always runs at a given speed (chosen by the game programmer)
			// regardless of how quickly the computer is able to execute the game's drawing/instructions.

			// CPS means Cycles Per Second. Higher CPS means faster gameplay.	GameRun() is called win_cps times per second.
			// FPS means Frames Per Second. Higher FPS means smoother gameplay. GameDraw() is called win_fps times per second.
			// Attaining the target CPS (game speed) is nearly guaranteed under normal circumstances, (< 10000 cps)
			// whereas attaining the target FPS (game smoothness) isn't always possible, depending on the game and cpu.

			// Warning: the following code is VERY "fragile" due to its time-sensitive nature, so even
			// the most innocent modification is likely to cause subtle problems if not drastic failure.

			// first check to see if the game changed the cps or fps and adjust accordingly
			if(cps != GameX.win_cps) {
				cps = GameX.win_cps;
				tpc = GetClocksPerSec() / cps;
				if(fps > cps) tpf = tpc;
				else tpf = GetClocksPerSec() / fps;
				update_auto_status = true;

			}
			if(fps != GameX.win_fps) {
				fps = GameX.win_fps;
				tpf = GetClocksPerSec() / fps;
				tpf = max(tpf,tpc);
				update_auto_status = true;
			}

			// sleep while draw update would exceed game update
			if(!needUpdate) AccurateSleep(tpc+cycleTime-GetClock());

			hyper thisTime = GetClock(); // variable to hold pre-action reference time

			int cpf_counter = 0; // number of times to update the game at this instant,
								 // if > 1 then we skip (cpf_counter-1) frames by updating again before drawing them

			// calculate how many times we need to update the game
			while(thisTime-cycleTime > tpc) {
				cycleTime += tpc;
				cpf_counter++;
			}
			// limit number of times to the maximum allowable updates per frame
			if(cpf_counter > GameX.win_max_cpf) {
				cpf_counter = GameX.win_max_cpf;
				frameTime = GetClock();
			}
/*
			// smooth out delay discrepencies between frames:
			// smoothing method 1:
			int new_todo = cpf_counter;
			int total_todo = new_todo + leftover;
			int trying = total_todo ? 1 + total_todo / 2 : 0;
			leftover = total_todo - trying;
			cpf_counter = trying;
*/
			// smoothing method 2:
			// (difficult to tell which is better -- method 1 is more correct in terms of achieving exactly the desired rate,
			//	but whatever rate it reaches sometimes appears slightly choppier than the same rate achieved with method 2)
			int ave = (cpf_counter+cpf_counter+cpf_counter + leftover) / 4;
			int total = cpf_counter + leftover;
			cpf_counter = total - ave; // takes into account the integer-divided nature of ave
			leftover = ave;

			if(cpf_counter) { // if we need to update the game state at all since last time:
				GameX.SetWarningOnDraw(true);
				while(cpf_counter--) { // update (without drawing) as many times as calculated
					GameX.UpdateInput(); // first update the input so GameRun() has fresh input data
					GameRun(); // call GameRun() function, whose contents decide the behaviour of the game
					cycle++; // increment counter so we can know how many cps we're really getting
					GameX.win_global_counter++;
					if(cycleTime > thisTime) { // correct overflow
						cycleTime = thisTime;
						cpf_counter = 0;
					}
				}
				needUpdate = true; // remember that the game screen now needs to be drawn again
				GameX.SetWarningOnDraw(false);
			}

			// sleep until the next draw update:
			AccurateSleep(tpf+frameTime-GetClock());

			// draw the game as necessary:
			thisTime = GetClock();
			if(thisTime-frameTime > tpf) { // executes at most once per loop
				if(needUpdate) {
					if(needUpdate && GameX.CanDraw()) {
						GameDraw(); // call GameDraw() function, whose contents draw the current game state
									// also draw game status if RUN_AUTOSHOWINFO is enabled:
						if(GameX.CanShowRunInfo()) {
							GameX.DrawStatus(update_auto_status);
							if(update_auto_status) update_auto_status = false;
						}
					}
					GameX.UpdateView(); // update onscreen view
					frame++; // increment counter so we can know how many fps we're really getting
					GameX.win_global_frame_total++;
					GameX.win_alternator = !GameX.win_alternator;
					frameTime = min(frameTime+tpf, thisTime);
					needUpdate = false;
				}
			} 

		#else
			// RUN_UNMANAGED (is defined):

			GameX.UpdateInput();
			GameMain();
			cycle++;
			frame++;
			GameX.win_global_counter++;
			GameX.win_global_frame_total++;

			if(GameX.CanShowRunInfo()) {
				GameX.DrawStatus(update_auto_status);
				if(update_auto_status) update_auto_status = false;
			}

			GameX.UpdateView();

		#endif

			hyper thetime = GetClock() - frameClock;// calculate the frame and cycle rate every second
			if(thetime >= GetClocksPerSec()) {		// so the game doesn't need to calculate it itself
				GameX.f_rate = frame;				// and can instead call GameX.GetCurrentFrameRate().
				GameX.c_rate = cycle;
				GameX.p_idle = (int)(100.0*(double)GameX.sleep_record/(double)GetClocksPerSec()); // calculate idle time percent (percent of power conserved)
				if(GameX.p_idle < 0 || GameX.p_idle > 99) GameX.p_idle = 99;
				frame = 0; cycle = 0; GameX.sleep_record = 0;
				frameClock = GetClock();
				frameTime = frameClock; // prevent long-term frame loss compensation
				update_auto_status = true;
			}
		}
	}
	if(GameX.win_log_debug)
		debug.Output("WinDX: Exiting normally...\n");

	GameX.CleanUp() ;
	return (int)(msg.wParam);
}


#ifndef GAMEX_RUN_UNMANAGED
void WindowsDX::SetGameSpeed (int cps)
{
	if(cps < 1) cps = 1;
	if(cps > 600) cps = 600;
	win_cps = cps;
}

void WindowsDX::SetLagLimit (int max_cpf)
{
	if(max_cpf==0) max_cpf = 0xFFFF; // 0 means no limit (for networked games)
	if(max_cpf<3) max_cpf = 3;
	win_max_cpf = max_cpf;
}

void WindowsDX::SetForcedFrameRate (int fps)
{
	win_fps = fps;
	if(fps>win_screenrefresh) win_vsync_enabled = false;
}

void WindowsDX::ResetForcedFrameRate (void)
{
	win_fps = -1;
}
#endif

int WindowsDX::UpdateView (void)
{
	if(win_status != GAMEX_READY)
		return 0;

	// end the scene here if it's still open
	if(draw_render_batch_open)
		EndRenderBatch();

	// DirectX 8.1
	//dd_device->Present (NULL, NULL, win_hwnd, NULL) ;

	// DirectDraw
	HRESULT status; 
	
	if(dd_front_buf->IsLost()) {
		dd_front_buf->Restore();

		dd_was_lost=1;

		if(win_log_debug)
			if(!dd_front_buf->IsLost())
				debug.Output("WinDX: DirectX surfaces were lost, but have now been properly restored.\n");

		if(dd_front_buf->IsLost())
			return 0;
		else {
			dd_main->RestoreAllSurfaces();

			// now we restore the 3D drawing device.
			// for some reason, if we don't do so RIGHT HERE
			// strange and inexplicable image corruption occurs

			UnlockBackBuffer() ;
			if(d3d_device != NULL) { 
				unsigned long temp_long = -1;
				status = d3d_device->ValidateDevice(&temp_long);

				if(status == DDERR_SURFACELOST) { // in newer versions of DirectX this is really D3DERR_DEVICELOST, which makes more sense
					d3d_device->Release();
					bool success = CreateDeviceForSurface(d3d_device, dd_back_buf); // may fail if UnlockBackBuffer() is not done before this
					d3d_device->SetRenderState(D3DRENDERSTATE_ALPHABLENDENABLE, TRUE); //Turn on alpha blending

					if(success) {
						if(win_log_debug)
							debug.Output("WinDX: The 3D rendering device has been lost and is now properly restored.\n");
					} else {
						if(win_err_debug)
							debug.Output("WinDX: The 3D rendering device has been lost and could NOT be restored.\n");
					}
				}
			}
		}
	}
	
	UnlockBackBuffer() ;

	// use Blt since we might want to scale the viewport to the screen
	// and, if in windowed mode, because we have a clipper set to the window

	// set the source and destination rectangles:
	RECT src_rect = {0, 0, win_width, win_height} ;
	RECT dst_rect = {win_xpos+win_xoffset, win_ypos+win_yoffset, win_xpos+win_xoffset+win_realx, win_ypos+win_yoffset+win_realy} ;

	if(win_vsync_enabled) {  // if vsync hasn't been disabled,
		SyncWithRefresh(); // wait until we're synchronized with the screen's refresh

		// debugging, delete the following code once vsync has been perfected:
		/*
				if(win_log_debug) {
				// this code draws a red line where the scanline is,
				// so the success of refresh synchronization can be seen
				// by how infrequently the red line is over the game
				// unfortunately the actual drawing takes a little time so the uncertainty principle applies
					DWORD scanline = -1;
					dd_main->GetScanLine(&scanline);
					int drawline = (scanline-(win_ypos+win_yoffset))*win_height/win_realy;
					if(drawline > 0 && drawline < win_height)
						InternalDrawRectDD(255,0,0, 0,drawline,win_width,drawline+2); // red line for after refreshsync
					UnlockBackBuffer() ;
				}
		*/
		// end of deletable debugging code
	}

	// do the blit:
	status = dd_front_buf->Blt(&dst_rect, dd_back_buf,
							   &src_rect,
							   DDBLT_WAIT,
							   NULL);

	// we might clear the back buffer at this point to avoid visual remnants of previous frames hanging around,
	// but it might be intentional in some circumstances for such remnants to be visible,
	// so, to allow greater freedom for the game developer, the back buffer is not automatically cleared.
	// (and the z-buffer clearing is handled elsewhere, in End3DScene)

	if(SUCCEEDED(status))
		return 1;
	else
		return 0;
}

// VSync method
void WindowsDX::SyncWithRefresh (void)
{
	// (the reason we don't use dd_main->WaitForVerticalBlank here
	//  is because it wastes a huge amount of CPU processing time.)

	DWORD scanline = 0; dd_main->GetScanLine(&scanline); // scanline position in pixels from top

	if((int)scanline <= win_ypos+win_yoffset || (int)scanline >= win_realy+win_ypos+win_yoffset)
		return; // return right away if we can already draw at this instant without tearing glitches

	hyper timecalc = (win_realy+win_ypos+win_yoffset - scanline) * (GetClocksPerSec() / CLOCKS_PER_SEC) / win_scanline_speed;

	AccurateSleep(timecalc,5); // sleep until we think we can draw without tearing glitches
							   // the 5 is to increase accuracy at the expense of a little processing power
}


void WindowsDX::DrawStatus(bool continuously)
{
	if(!continuously && f_rate!=0 && win_global_frame_total%(f_rate/((win_fullscreen)?4:1))!=0)
		return;

	if(!win_request_no_mouse_input && (continuously || (f_rate!=0 && win_global_frame_total%f_rate == 0))) {
		di_statMouseX = GameX.GetMouseX();
		di_statMouseY = GameX.GetMouseY();
	}

	char mouseStr [64];
	if(win_request_no_mouse_input) sprintf(mouseStr,"");
	else sprintf(mouseStr," - (%d, %d) mouse",di_statMouseX,di_statMouseY);

	char str [256];
	sprintf(str,"%d/%d fps - %d/%d cps - %d/%d idle%s",GetCurrentFrameRate(),GetTargetFrameRate(),GetCurrentCycleRate(),GetTargetCycleRate(),GetCurrentIdlePercent(),100,mouseStr);

	if(!win_fullscreen)
		SetWindowText(win_hwnd,str);
	else
		DrawOutlinedText(12,win_height-24,str);
}
